#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <cmath>
#include <sstream>
#include <csignal>
#include <vector>
#include <cstdlib>
#include <TROOT.h>
#include <TCanvas.h>
#include <TProfile2D.h>
#include <TF1.h>
#include <TH1.h>
#include "AsciiRoot.h"
#include "utils.h"
#include "Tracer.h"
#include "Hit.h"

bool _A_do_run = true;
void _A_got_intr(int)
{
    _A_do_run = false;
}

// decodes the header and returns a vector with the integers found
std::vector<int> decode_header(const std::string &h)
{
    std::vector<int> vout;
    std::istringstream istr(h);
    char buf[256];
    while (istr)
    {
        istr.getline(buf, sizeof(buf), ';');
        if (!istr)
            break;

        vout.push_back(atoi(buf) );
    }
    return vout;
}

AsciiRoot::AsciiRoot(const char *nam, const char *pedfile, const char *gainfile) :
    _seedcut(5.), _neighcut(2.5), _average_gain(1.)
{
    int i;
    for (i=0;i<256;i++)
    {
        _ped[i] = 0.;
        _gain[i] = 1.;
        _noise[i] = 1.;
        _mask[i] = false;
    }
    if (nam)
        open(nam);
    
    if (pedfile)
        load_pedestals(pedfile);
    
    if (gainfile)
        load_gain(gainfile);
}
AsciiRoot::~AsciiRoot()
{
    if (ifile)
    {
        ifile->close();
    }
}
void AsciiRoot::open(const char *name)
{
    ifile = new std::ifstream(name);
    if (!(*ifile))
    {
        std::cout << "Could not open data file: " << name << std::endl;
        delete ifile;
        ifile = 0;
		return;
    }
    std::string header;
    unsigned int ic, lheader;
    char c;
    ifile->read((char *)&_t0, sizeof(time_t));
    ifile->read((char *)&_type, sizeof(int));
    ifile->read((char *)&lheader, sizeof(unsigned int));
    for (ic=0; ic<80; ic++)
    {
        ifile->read(&c, sizeof(char));
        header.append(1, c);
    }
    header = trim_str(header);

    std::cout << "type: " << _type << " header: " << header << std::endl;
    std::vector<int> param = decode_header(header);
    ifile->read((char *)_ped, 256*sizeof(double));
    ifile->read((char *)_noise, 256*sizeof(double));
    switch (_type)
    {
        case 1: // calibration
        case 2: // laser sync
            _npoints = param[0];
            _from = param[1];
            _to = param[2];
            _step = param[3];
            break;
        case 3: // laser run
        case 4: // source run
            _nevts = param[0];
            _npoints = _from = _to = _step = 0;
            break;
    }
    data_start = ifile->tellg();
}

void AsciiRoot::rewind()
{
    if (ifile)
    {
        ifile->clear();
        ifile->seekg(data_start, std::ios::beg);
    }
}

void AsciiRoot::close()
{
    if (ifile)
    {
        ifile->close();
        delete ifile;
        ifile = 0;
    }
}

int AsciiRoot::read_event()
{
    if (ifile)
    {
        ifile->read((char *)&_data, sizeof(EventData));
        if (ifile->eof())
        {
            std::cout << "End of file" << std::endl;
            return -1;
        }
        else
        {
            //process_event();
            return 0;
        }
    }
    else
        return -1;
}


double AsciiRoot::time() const
{
    unsigned short fpart = _data.time & 0xffff;
    short ipart = (_data.time & 0xffff0000)>>16;
    if (ipart<0)
        fpart *= -1;
    //double tt = 100.*(1. -(ipart + (fpart/65535.)));
    double tt = 100.0*(ipart + (fpart/65535.));
    return tt;
}

double AsciiRoot::temp() const
{
    return 0.12*_data.temp - 39.8;
}

TH1 *AsciiRoot::show_pedestals()
{
    int ic;
    TH1 *hst = new TH1D("hPed","Pedestals",256,-0.5, 255.5);
    hst->SetYTitle("ADCs");
    hst->SetXTitle("Channel no.");
    for (ic=0; ic<256; ic++)
        hst->SetBinContent(ic+1, _ped[ic]);

    return hst;
}

TH1 *AsciiRoot::show_noise()
{
    int ic;
    TH1 *hst = new TH1D("hNoise","Noise",256,-0.5, 255.5);
    if (gain()==1)
    {
        hst->SetYTitle("ADCs");
    }
    else
    {
        hst->SetYTitle("e^{-} ENC");
    }
    hst->SetXTitle("Channel no.");
    for (ic=0; ic<256; ic++)
        hst->SetBinContent(ic+1, noise(ic));

    return hst;
}

void AsciiRoot::compute_pedestals_fast(int mxevts, double wped, double wnoise)
{
    int i, ievt;
    
    if (!ifile)
        return;

    if (mxevts<0)
        mxevts = 100000000;

    for (i=0;i<256;i++)
        _ped[i] = _noise[i] = 0.;
    
    std::ifstream::pos_type here = ifile->tellg();
    std::cout << "Computing fast pedestas..." << std::endl;
    for (ievt=0; read_event()==0 && ievt<mxevts; ievt++)
    {
        if (!(ievt%100))
        {
            std::cout << "\revent " << std::setw(10) << ievt << std::flush;
        }
        common_mode();
        for (i=0; i<256; i++)
        {
            int ichip = i/128;
            // IF noise is 0, set it arbitrarily to 1.
            if (_noise[i]==0.)
                _noise[i] = 1.;

            if (_ped[i]==0.)
            {
                // If pedestal is not yet computed we assume the current
                // channel value should not be too far
                _ped[i] = _data.data[i];
            }
            else
            {
                // Do the pedestal and noise correction
                double corr;
                double xs;

                _signal[i] = _data.data[i] - _ped[i];
                corr = _signal[i] * wped;
                
                xs = (_signal[i]-_cmmd[ichip])/_noise[i];
                if (corr > 1.)
                    corr = 1.;

                if (corr < -1)
                    corr = -1.;

                _ped[i] += corr;

                if (fabs(xs) < 3.)
                {
                    _noise[i] = _noise[i]*(1.0-wnoise) + xs*xs*wnoise;
                }
            }
        }
    }
    std::cout << "\nDone" << std::endl;
    rewind();    
}

TH2 *AsciiRoot::compute_pedestals(int mxevts, bool do_cmmd)
{
    if (!ifile)
        return 0;

    if (mxevts<0)
        mxevts = 100000000;

    int ievt, ichan;
    TH2 *hst = create_h2("hRaw","Raw data",256, -0.5,255.5,256, -0.5,1023.5);
    TH2 *hsts = create_h2("hSig","Signal",256, -0.5,255.5,256, -127.5,127.5);
    

    std::ifstream::pos_type here = ifile->tellg();
    std::cout << "Computing pedestas..." << std::endl;
    for (ievt=0; read_event()==0 && ievt<mxevts; ievt++)
    {
        process_event(do_cmmd);
        for (ichan=0; ichan<256; ichan++)
            hst->Fill(ichan, data(ichan)-get_cmmd(ichan/128));
        
        if (!(ievt%100))
        {
            std::cout << "\revent " << std::setw(10) << ievt << std::flush;
        }
    }
    std::cout << "\nDone" << std::endl;
    rewind();

    for (ichan=0; ichan<256; ichan++)
    {
        TF1 *g = new TF1("g1", "gaus");
        TH1 *h1 = hst->ProjectionY("__hx__", ichan+1, ichan+1);
        g->SetParameters(h1->GetSumOfWeights(), h1->GetMean(), h1->GetRMS());
        g->SetRange(h1->GetMean()-2.5*h1->GetRMS(), h1->GetMean()+2.5*h1->GetRMS());
        h1->Fit("g1", "q0wr");
        _ped[ichan] = h1->GetFunction("g1")->GetParameter(1);
        _noise[ichan] = h1->GetFunction("g1")->GetParameter(2);
        delete h1;
        delete g;
    }
    
    rewind();
    for (ievt=0; read_event()==0 && ievt<mxevts; ievt++)
    {
        process_event(do_cmmd);
        for (ichan=0; ichan<256; ichan++)
            hsts->Fill(ichan, signal(ichan));

        if (!(ievt%100))
        {
            std::cout << "\revent " << std::setw(10) << ievt << std::flush;
        }
    }
    std::cout << "\nDone" << std::endl;
    rewind();    
    
    return hst;
}

void AsciiRoot::find_clusters(double polarity, int ichip)
{
    int i, imax=-1, left, right;
    double mxsig=-1.e20, sg, val;
    bool used[256];
    int chan0=0;
    int chan1=256;
    
    if (ichip>=0 && ichip<2)
    {
        chan0 = ichip*128;
        chan1 = (ichip+1)*128;
    }
    
    
    for (i=0;i<256;i++)
    {
        used[i]= _mask[i] ? true : false;
    }
    clear();

        
    while (true)
    {
        /*
         * Find the highest 
         */
        imax = -1;
        for (i=chan0; i<chan1; i++)
        {
            if (used[i] || _signal[i]*polarity<0.)
                continue;
            
            if (fabs(sn(i))>_seedcut)
            {
                val = fabs(signal(i));
                if (mxsig<val)
                {
                    mxsig = val;
                    imax = i;
                }
            }
        }

        if (imax<0)
            break;

        sg = signal(imax);
        used[imax]=true;
        // Now look at neighbours
        // first to the left
        left = imax;
        for (i=imax-1;i>=0;i--)
        {
            if ( _signal[i]*polarity<0.)
                break;
            
            if ( fabs(sn(i)) > _neighcut )
            {
                used[i] = true;
                sg += signal(i);
                left = i;
            }
            else
		// TODO: this needs to be removed
        // The idea is to merge to clusters that hanve only one strip in between
	    // In the laser runs this is a consequences of reflections...
	    {
	       int j = i-1;
	       if (j>=0  )
	       {
		  if ( fabs(sn(j)) > _neighcut )
		     continue;
	       }
	       break;
	    }
        }
        // now to the right
        right = imax;
        for (i=imax+1;i<256;i++)
        {
            if ( _signal[i]*polarity<0.)
                break;
            if ( fabs(sn(i))>_neighcut )
            {
                used[i] = true;
                sg += signal(i);
                right = i;
            }
            else
		// TODO: this needs to be removed
        // The idea is to merge to clusters that hanve only one strip in between
	    // In the laser runs this is a consequences of reflections...
	    {
	       int j = i+1;
	       if (j<256)
	       {
		  if ( fabs(sn(j)) > _neighcut )
		     continue;
	       }
                break;
	    }
        }
        add_hit(Hit(imax, left, right, sg));
    }
}

void AsciiRoot::save_pedestals(const char *fnam)
{
    std::ofstream ofile(fnam);
    if (!ofile)
    {
        std::cout << "Could not open " << fnam << " to save pedestals." << std::endl;
        return;
    }
    int i;
    for (i=0; i<256; i++)
    {
        ofile << _ped[i] << "\t" << _noise[i] << "\n";
    }
    ofile.close();
}

void AsciiRoot::load_pedestals(const char *fnam)
{
    std::ifstream ifile(fnam);
    if (!ifile)
    {
        std::cout << "Could not open " << fnam << " to load pedestals." << std::endl;
        return;
    }
    int i;
    for (i=0; i<256; i++)
    {
        ifile >> _ped[i] >> std::ws >> _noise[i] >> std::ws;
        _mask[i] = (_noise[1]>20. || _noise[i]<=0.);
    }
    ifile.close();
}

void AsciiRoot::load_gain(const char *fnam)
{
    std::ifstream ifile(fnam);
    if (!ifile)
    {
        std::cout << "Could not open " << fnam << " to load the gain." << std::endl;
        return;
    }
    int i;
    int ichan;
    double val, xn, xm;
    xn=xm=0.;
    for (i=0; i<256; i++)
    {
        ifile >> ichan >> std::ws;
        if (ifile.eof())
            break;
        ifile >> val;
        if (ifile.eof())
            break;

        xn++;
        xm += val;
        _gain[ichan] = val;
        
        ifile >> std::ws;
        if (ifile.eof())
            break;
    }
    if (xn>0)
    {
        _average_gain = xm/xn;
    }
    ifile.close();
}

void AsciiRoot::process_event(bool do_cmmd)
{
    int i;
    for (i=0; i<256; i++)
    {
        _signal[i] = _data.data[i]-_ped[i];
        _sn[i] = _noise[i]>1. && !_mask[i] ? _signal[i]/_noise[i] : 0.;
    }
    if (do_cmmd)
    {
        int ichip=-1;
        common_mode();    

        for (i=0; i<256; i++)
        {
            if (!(i%128))
                ichip ++;
                
            _signal[i] = _data.data[i]-_ped[i] - _cmmd[ichip];
            _sn[i] = (_noise[i] >1. && !_mask[i] ? _signal[i]/_noise[i] : 0.);
        }
    }
}
void AsciiRoot::common_mode()
{
    int ip, i, j;

    
    for (j=0; j<2; j++)
    {
        int offs = 128*j;
        int sstop = 128*(j+1);
        double mean, sm, xn, xx, xs, xm, tmp;
        bool use_it;
        mean = sm = 0.;
        //std::cout << "Chip " << j << std::endl;    
        for (ip=0;ip<3;ip++)
        {
            xn = xs = xm = 0.;
            for (i=offs; i<sstop; i++)
            {
                if (_mask[i])
                    continue;
                
                use_it = true;
                xx = data(i) - _ped[i];
                if (ip)
                {
                    tmp = fabs((xx-mean)/sm);
                    use_it = (tmp<2.5);
                }
                if (use_it)
                {
                    xn++;
                    xm += xx;
                    xs += xx * xx;
                }
            }
            if (xn>0.)
            {
                mean = xm / xn;
                sm = sqrt( xs/xn - mean*mean);
            }
          //  std::cout << "...iter " << ip << ": xm " << mean << " xs: " << sm << std::endl;
        }
        _cmmd[j] = mean;
        _cnoise[j] = sm;
        
    }
}

void AsciiRoot::spy_data(bool with_signal, int nevt)
{
    TVirtualPad *pad;
    if (!ifile)
        return;
    
    sighandler_t old_handler = ::signal(SIGINT, _A_got_intr);
    _A_do_run = true;

    TCanvas *cnvs = (TCanvas *)gROOT->FindObject("cnvs");
    if (cnvs)
    {
        cnvs->Clear();
    }
    else
       cnvs = new TCanvas("cnvs","cnvs", 700, 800);
    
    cnvs->Divide(2,3);
    
    
    TH1 *hsignal = create_h1("hsignal","signal (ADC)",256, -0.5, 255.0);
    hsignal->SetXTitle("Channel");
    hsignal->SetYTitle("ADC");
    hsignal->SetMinimum(-150);
    hsignal->SetMaximum(150);

    TH1 *helec = create_h1("helec","signal (elec)", 256, -0.5, 255.5);
    helec->SetXTitle("Channel");
    helec->SetYTitle("electrons");
    helec->SetMinimum(-150/gain());
    helec->SetMaximum(150/gain());
    
    TH1 *hraw = create_h1("hraw","Raw Data (around 512.)",256, 0., 256.);
    hraw->SetXTitle("Channel");
    hraw->SetYTitle("ADC");
    hraw->SetMinimum(-150);
    hraw->SetMaximum(+150);
    
    TH1 *hrawc = create_h1("hrawc","Raw Data (no commd)",256, 0., 256.);
    hrawc->SetXTitle("Channel");
    hrawc->SetYTitle("ADC");
    hrawc->SetMinimum(-150);
    hrawc->SetMaximum(+150);
    

    TH1 *hcmmd[2];
    hcmmd[0] = create_h1("hcmmd0","Common mode (Chip 0)",50,-100.,100.);
    hcmmd[0]->SetXTitle("Common mode");
    hcmmd[1] = create_h1("hcmmd1","Common mode (Chip 1)",50,-100.,100.);
    hcmmd[1]->SetXTitle("Common mode");
    
    int ievt,jevt;
    for (ievt=jevt=0; read_event()==0 && _A_do_run && ievt<nevt;jevt++)
    {
        process_event();
        find_clusters();
        if ( with_signal && empty())
            continue;
        
        int i,ichip=-1;
        for (i=0; i<256; i++)
        {
            if (!(i%128))
                ichip++;
            
            hsignal->SetBinContent(i+1, _signal[i]);
            helec->SetBinContent(i+1, signal(i));
            hraw->SetBinContent(i+1,data(i)-512.);
            hrawc->SetBinContent(i+1, data(i)-_ped[i]);
            hcmmd[ichip]->Fill(_signal[i]+get_cmmd(ichip));
        }
        pad = cnvs->cd(1);
        pad->SetGrid(1,1);
        hsignal->Draw();
        pad = cnvs->cd(2);
        pad->SetGrid(1,1);
        helec->Draw();
        
        pad = cnvs->cd(3);
        pad->SetGrid(1,1);
        hraw->Draw();
        
        pad = cnvs->cd(4);
        pad->SetGrid(1,1);
        hrawc->Draw();

        pad = cnvs->cd(5);
        pad->SetGrid(1,1);
        hcmmd[0]->Draw();

        pad = cnvs->cd(6);
        pad->SetGrid(1,1);
        hcmmd[1]->Draw();
            
        std::cout << std::setiosflags(std::ios::fixed);
        std::cout << "*** Event " << jevt << " *****" << std::endl;
        std::cout << "Common Mode:" << std::endl
                  << "   Chip 0 " << std::setw(6) << std::setprecision(1) << get_cmmd(0) << " noise: " << get_cnoise(0)
                  << std::endl
                  << "   Chip 1 " << std::setw(6) << std::setprecision(1) << get_cmmd(1) << " noise: " << get_cnoise(1)
                  << std::endl;   
                  
        std::cout << "Time: " << time() << " ns" << std::endl;
        std::cout << "Clusters: " << std::endl;

        AsciiRoot::HitList::iterator ip;
        for (ip=begin(); ip!=end(); ++ip)
        {
            std::cout << "   chan: " << ip->center() 
                      << " sig: "
                      << std::setw(6) << std::setprecision(1) << ip->signal()
                      << " left: " << ip->left() << " right: " << ip->right()
                      << std::endl;
            std::cout << '\t' << "channels: " << std::endl;
            int j;
            for (j=ip->left();j<=ip->right();j++)
                std::cout << "\t   " << j << " sn: " << _sn[j] << " signal: " << _signal[j] << " noise: " << _noise[j] << '\n';
            std::cout << std::endl;
        }
        
        cnvs->Update();
        ievt++;
    }
    std::cout << std::endl;
    _A_do_run= true;
    ::signal(SIGINT, old_handler);
}

