#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdlib>
#include <TCanvas.h>
#include <TROOT.h>
#include <TDirectory.h>
#include <TStyle.h>
#include <TH1.h>
#include <TH2.h>
#include <TF1.h>
#include <TRegexp.h>
#include <TSystem.h>
#include <TGraphErrors.h>
#include "AsciiRoot.h"
#include "analysis.h"

AsciiRoot *gAsciiRoot=0;

void sin_preguntas(const char *data_file, const char *cal_file=0, const char *ped_file=0,
                   int polarity=0, bool dofit=true)
{
    const char *ped_f = "/tmp/alibava_ped.ped";
    const char *cal_f = "/tmp/alibava_cal.cal";

    // Prepare the file
    if (gAsciiRoot)
        delete gAsciiRoot;

    AsciiRoot *A;

    // Compute the pedestals
    if (!ped_file)
    {
        std::cout << "*** Computing pedestals\n" 
                  << "***********************" << std::endl;
        A = new AsciiRoot(data_file,0);
        A->compute_pedestals_fast();
        A->save_pedestals(ped_f);
        {
            TCanvas *pcnvs = new TCanvas("pcnvs","Pedestals and Noise");
            pcnvs->Divide(1, 2);
            pcnvs->cd(1);
            A->show_pedestals()->Draw();
            pcnvs->cd(2);
            A->show_noise()->Draw();
            pcnvs->Update();
        }
        delete A;
    }
    else
        ped_f = ped_file;

    // Get calibratioin file
    if (cal_file)
    {
        std::cout << "*** Computing calibration factors" << std::endl
                << "*********************************" << std::endl;
        AsciiRoot *B = new AsciiRoot(cal_file,ped_f);
        analyze(*B, "hGain");
        save_text_file((TH1 *)gDirectory->Get("hGain"), cal_f);
        TCanvas *cnvs = (TCanvas *)gROOT->FindObject("Gain");
        if (cnvs)
            delete cnvs;

        cnvs = new TCanvas("Gain","Gain",700,800);
        cnvs->Divide(1, 2);
        cnvs->cd(1)->SetGrid(1, 1);
        gDirectory->Get("hGain")->Draw();
        cnvs->cd(2)->SetGrid(1, 1);
        gDirectory->Get("hGain_all")->Draw("surf3z");
        TCanvas *c5 = new TCanvas("c5","Projection",600,800,400,300);
        c5->SetGrid(1, 1);
        gDirectory->Get("hGain_py")->Draw("L");
        delete B;
    }
    else
        cal_f = 0;

    // Analyze the data file
    std::cout << "*** Analyzing the data file" << std::endl
            << "***************************" << std::endl;
    A = new AsciiRoot(data_file,ped_f, cal_f);
    gAsciiRoot = A;
    A->set_cuts(6., 2.5);
    analyze(*A, "hSpec", polarity);
    if (A->type()>2)
    {
        double t1, t2;
        if (A->type()==3) // Laser run
        {
            t1 = 0.;
            t2 = 1000.;
            dofit = false;
        }
        else // RadSource
        {
            t1 = 12.;
            t2 = 22.;
        }
        show_spectrum("hSpec", t1, t2, polarity, A, dofit);
    }
    else if (A->type()==2)
    {
        int ix, iy, iz;
        TVirtualPad *pad;
        TCanvas *cnvs = (TCanvas *)gROOT->FindObject("cnvs");
        if (cnvs)
            delete cnvs;

        gStyle->SetOptStat(0);
        cnvs = new TCanvas("cnvs","Laser Delay", 700,800);
        cnvs->Divide(1, 2);
        pad = cnvs->cd(1);
        pad->SetGrid(1, 1);
        pad->SetObjectStat(false);
        gDirectory->Get("hSpec")->Draw();

        pad = cnvs->cd(2);
        pad->SetGrid(1, 1);
        pad->SetObjectStat(false);
        TH2 *h2 = (TH2 *)gDirectory->Get("hSpec_all");
        h2->GetMaximumBin(ix, iy, iz);
        h2->GetXaxis()->SetRange(ix-5, ix+5);
        h2->Draw("surf4");
    }

}

double get_volt(const char *fnam)
{
    char *str =strdup(fnam);
    char *q, *p = strstr(str, "laser_");
    if (!p)
    {
        free(str);
        return 0.;
    }

    std::cout << "**** fname: p0 " << p;
    p+= strlen("laser_");
    q = strchr(p, '_');
    if (!q)
    {
        q = strchr(p, '.');
        if (!q)
            return 0.;
    }

    *q=0;
    std::cout << " p1 " << p << std::endl;
    free(str);
    return atof(p);
}

TGraphErrors *scan_laser(const char *detector, const char *cal_file=0,
                         const char *ped_file=0, int polarity=0, bool dofit=true,
                         int mxevts=-1)
{
    const char *ped_f = "/tmp/alibava_ped.ped";
    const char *cal_f = "/tmp/alibava_cal.cal";

    /*
     * Compute pedestals.
     */
    if (!ped_file)
    {
        std::cout << "I need a pedestal file for laser data" << std::endl;
        return 0;
    }
    else
        ped_f = ped_file;

    if (gAsciiRoot)
        delete gAsciiRoot;

    AsciiRoot *A;

    /*
     * Get calibration file
     */
    if (cal_file)
    {
        std::cout << "*** Computing calibration factors" << std::endl
                << "*********************************" << std::endl;
        AsciiRoot *B = new AsciiRoot(cal_file,ped_f);
        analyze(*B, "hGain");
        save_text_file((TH1 *)gDirectory->Get("hGain"), cal_f);
        delete B;
    }
    else
        cal_f = 0;

    /*
     * Scan the laser files
     */
    // Analyze the data file
    std::cout << "*** Analyzing the data files for " << detector << std::endl
            << "***************************" << std::endl;

    TGraphErrors *gout = new TGraphErrors();
    gout->SetName(detector);
    int npts;
    char *str, path[256];
    char fnam[256];
    sprintf(path, "ls -1 %s/Analysis_Hamamatsu/%s/laser_*.dat", getenv("PWD"), detector);
    FILE *pipe = gSystem->OpenPipe(path, "r");
    for (npts=0, str=fgets(fnam, sizeof(fnam), pipe); str; str=fgets(fnam, sizeof(fnam),
                                                                     pipe))
    {
        *(strchr(fnam, '\n'))=0;
        TString *sf = new TString(fnam);
        std::cout << fnam << std::endl;
        double volt = get_volt(fnam);

        A = new AsciiRoot(fnam,ped_f, cal_f);
        gAsciiRoot = A;
        A->set_cuts(6., 2.5);
        analyze(*A, "hSpec", polarity, -1, 0, mxevts);
        show_spectrum("hSpec", 0., 1000., polarity, A, dofit);

        TH1 *hst = (TH1 *)gDirectory->Get("hSpec_timecut");
        TF1 *func = hst->GetFunction("gaus");
        if (func)
        {
            double mean = func->GetParameter(1);
            double sig = func->GetParError(1) ;
            std::cout << volt << " " << mean << ", " << sig << ", "
                    << func->GetParameter(2) << std::endl;

            gout->SetPoint(npts, volt, mean);
            gout->SetPointError(npts, 0., sig);
            npts++;
        }

        delete sf;
    }
    gSystem->ClosePipe(pipe);
    return gout;
}

void graph2csv(TGraphErrors *gr, const char *ofilename=0)
{
    char fnam[256];
    if (!ofilename)
        sprintf(fnam, "%s.csv", gr->GetName());
    else
        strcpy(fnam, ofilename);

    std::ofstream ofile(fnam);

    ofile << "Volt\tQ\teQ" << std::endl;
    int ip, np=gr->GetN();
    for (ip=0; ip<np; ip++)
    {
        ofile << gr->GetX()[ip] << '\t' << gr->GetY()[ip] << '\t' << gr->GetEY()[ip] << std::endl;
    }

    ofile.close();
}

int printCanvas(TCanvas *cnvs, const char *out, int res=300)
{
    int rc = 0;
    char line[512];
    cnvs->SaveAs("/tmp/cnvs.eps");
    sprintf(line, "convert -units PixelsPerInch -density %d /tmp/cnvs.eps %s", res, out);
    rc = gSystem->Exec(line);
    gSystem->Unlink("/tmp/cnvs.eps");
    return rc;
}
